#include "Renderer.h"
#include "Components/SpriteComponent.h"
#include "utils/Shader.h"

#include <GL/glew.h>
#include <glm/glm.hpp>
#include <iostream>

static Shader spriteShader;

GLuint Renderer::fbo = 0;
GLuint Renderer::textureColorBuffer = 0;
GLuint Renderer::rbo = 0;
GLuint Renderer::quadVAO = 0;
GLuint Renderer::quadVBO = 0;
int Renderer::width = 1280;
int Renderer::height = 720;

void Renderer::InitQuad() {
    float vertices[] = {
        // pos    // uv
        0.f, 0.f, 0.f, 0.f,
        1.f, 0.f, 1.f, 0.f,
        1.f, 1.f, 1.f, 1.f,
        0.f, 1.f, 0.f, 1.f
    };

    glGenVertexArrays(1, &quadVAO);
    glBindVertexArray(quadVAO);

    glGenBuffers(1, &quadVBO);
    glBindBuffer(GL_ARRAY_BUFFER, quadVBO);
    glBufferData(GL_ARRAY_BUFFER, sizeof(vertices), vertices, GL_STATIC_DRAW);

    glEnableVertexAttribArray(0); // position
    glVertexAttribPointer(0, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), (void*)0);
    glEnableVertexAttribArray(1); // UV
    glVertexAttribPointer(1, 2, GL_FLOAT, GL_FALSE, 4 * sizeof(float), (void*)(2 * sizeof(float)));

    glBindVertexArray(0);
}

void Renderer::Init() {
    glGenFramebuffers(1, &fbo);
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);

    glGenTextures(1, &textureColorBuffer);
    glBindTexture(GL_TEXTURE_2D, textureColorBuffer);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, nullptr);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glFramebufferTexture2D(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_TEXTURE_2D, textureColorBuffer, 0);

    glGenRenderbuffers(1, &rbo);
    glBindRenderbuffer(GL_RENDERBUFFER, rbo);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH24_STENCIL8, width, height);
    glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_DEPTH_STENCIL_ATTACHMENT, GL_RENDERBUFFER, rbo);

    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
        std::cerr << "Framebuffer is not complete!" << std::endl;

    glBindFramebuffer(GL_FRAMEBUFFER, 0);

    InitQuad();

    spriteShader.LoadFromFile("src/assets/shaders/sprite.vert", "src/assets/shaders/sprite.frag");
}

void Renderer::Resize(int w, int h) {
    if (w == width && h == height) return;
    width = w;
    height = h;

    glBindTexture(GL_TEXTURE_2D, textureColorBuffer);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, nullptr);

    glBindRenderbuffer(GL_RENDERBUFFER, rbo);
    glRenderbufferStorage(GL_RENDERBUFFER, GL_DEPTH24_STENCIL8, width, height);
}

void Renderer::Begin() {
    glBindFramebuffer(GL_FRAMEBUFFER, fbo);
    glViewport(0, 0, width, height);
    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);
}

void Renderer::End() {
    glBindFramebuffer(GL_FRAMEBUFFER, 0);
}

//void Renderer::DrawSprite(SpriteComponent* sprite, const glm::vec2& pos, float scale) {
//    GLuint tex = sprite->GetTextureID();
//    if (!tex) return;
//
//    glBindTexture(GL_TEXTURE_2D, tex);
//    glBegin(GL_QUADS);
//    float size = 100.0f * scale;
//
//    glTexCoord2f(0, 0); glVertex2f(pos.x, pos.y);
//    glTexCoord2f(1, 0); glVertex2f(pos.x + size, pos.y);
//    glTexCoord2f(1, 1); glVertex2f(pos.x + size, pos.y + size);
//    glTexCoord2f(0, 1); glVertex2f(pos.x, pos.y + size);
//    glEnd();
//}




void Renderer::DrawSprite(SpriteComponent* sprite, const glm::vec2& pos, float zoom, glm::vec2& CameraPos) {
    if (!sprite || sprite->GetTextureID() == 0) return;

    spriteShader.Use();
    glm::vec2 size = sprite->GetSize();
    glm::vec2 screenPos = (pos - CameraPos) * zoom + glm::vec2(size.x / 2, size.y / 2);
    spriteShader.SetVec2("uPos", screenPos);
    spriteShader.SetVec2("uSize", size*zoom);
    spriteShader.SetVec2("uScreen", glm::vec2(width, height));
    spriteShader.SetInt("uTex", 0);

    glActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, sprite->GetTextureID());

    glBindVertexArray(quadVAO);
    glDrawArrays(GL_TRIANGLE_FAN, 0, 4);
    glBindVertexArray(0);
}


void Renderer::DrawEditorGrid(const glm::vec2& cameraPos, float zoom) {
    glUseProgram(0);
    glColor4f(0.5f, 0.5f, 0.5f, 0.25f);
    glLineWidth(1.0f);

    glBegin(GL_LINES);
    float spacing = 100.0f;
    float viewWidth = width / zoom;
    float viewHeight = height / zoom;

    float left   = cameraPos.x - viewWidth / 2;
    float right  = cameraPos.x + viewWidth / 2;
    float bottom = cameraPos.y + viewHeight / 2;
    float top    = cameraPos.y - viewHeight / 2;

    int minX = static_cast<int>(std::floor(left / spacing)) * spacing;
    int maxX = static_cast<int>(std::ceil(right / spacing)) * spacing;
    int minY = static_cast<int>(std::floor(top / spacing)) * spacing;
    int maxY = static_cast<int>(std::ceil(bottom / spacing)) * spacing;

    for (int x = minX; x <= maxX; x += spacing) {
        glVertex2f((float)x, top);
        glVertex2f((float)x, bottom);
    }

    for (int y = minY; y <= maxY; y += spacing) {
        glVertex2f(left, (float)y);
        glVertex2f(right, (float)y);
    }

    glEnd();
}


GLuint Renderer::GetRenderTexture() {
    return textureColorBuffer;
}

glm::ivec2 Renderer::GetSize() {
    return { width, height };
}
